import typer
import base64

cli = typer.Typer()


@cli.command()
def encode_json_header(
        json: str = typer.Option(..., prompt=True, help="JSON string to encrypt"),
):
    json_bytes = json.encode('utf-8')
    base64_bytes = base64.b64encode(json_bytes)
    base64_str = base64_bytes.decode('utf-8')
    typer.echo(base64_str)



from typing import List, Dict
from ehelply_bootstrapper.utils.state import State

import typer
import asyncio

cli = typer.Typer()

"""
CONSTANTS
"""
RESERVED_WORDS: List[str] = ["playground", "test", "testing", "privileged"]
EHELPLY_HEADERS: str = "ehelply-active-participant,ehelply-project,ehelply-data,sentry-trace"
WEB_ACL_ARN: str = "arn:aws:wafv2:ca-central-1:330779307994:regional/webacl/ehelply-test/5d87bb76-a08c-4136-a875-f6f5f89e4dd2"


@cli.command()
def sync_api_gateway(
        include_service_routers: bool = typer.Option(False, help="Include service routers in the spec"),
        ignore_prefixes: int = typer.Option(1,
                                            help="How many prefixes in the path to ignore. Typically you will want to set this to 1 (and it is 1 by default) unless you are syncing a service with multiple root routers. In that case, set this to 0."),
        target_group_port: str = typer.Option(
            ..., prompt=True
        ),
):
    """Script to export Open API JSON spec to a file. Exports to /api-specs"""

    typer.echo()
    typer.echo("=========================================")
    typer.echo(" CONFIRM PARAMETERS:")
    typer.echo(f" Target Group Port: {target_group_port}")
    typer.echo(f" Ignore Prefixes: {ignore_prefixes}")
    typer.echo()
    typer.echo(
        " API Gateway Sync'ing is a destructive process. Your ENTIRE API Gateway API will be overwritten while doing this. Please ensure that you intend to do this.")
    typer.echo()
    typer.echo(
        " Quick explanation of 'ignore_prefixes': How many prefixes in the path to ignore. Typically you will want to set this to 1 (and it is 1 by default) unless you are syncing a service with multiple root routers. In that case, set this to 0.")
    typer.echo("=========================================")
    typer.echo()

    confirm = typer.confirm("Are you sure you want to continue with these parameters?")
    if not confirm:
        raise typer.Abort()

    with typer.progressbar(length=8, label="Syncing") as progress:
        typer.echo("Starting API Gateway Sync. Expect to see lots of bootstrap text in the log")

        import json

        import sys

        from pathlib import Path

        from datetime import datetime

        from ehelply_microservice_library.cli.cli_state import CLIState

        class DocsService(CLIState.service):
            def if_dev_launch_dev_server(self) -> bool:
                return False

            def after_fast_api(self):
                asyncio.run(super().startup_event())

            def is_load_service_routers(self) -> bool:
                return include_service_routers

        progress.update(1)
        typer.echo()

        service = DocsService()

        typer.echo()
        progress.update(2)
        typer.echo()

        if service.fastapi_driver:
            app = service.fastapi_driver.instance

            service_meta = service.service_meta
            api_spec: dict = app.openapi()

            # Remove "eHelply" from the name.
            temp_name = service_meta.name.split(" ")
            temp_name.remove("eHelply")

            api_name: str = "".join(temp_name) + " (AutoGen)"

            apigateway_client = State.aws.make_client(name='apigateway')

            apigateway: dict = None

            # Get API Gateway if it exists
            for apigateway_api in apigateway_client.get_rest_apis(limit=500)['items']:
                if api_name == apigateway_api['name']:
                    apigateway = apigateway_api

            typer.echo()

            # Make a new API in API Gateway if it doesn't exist
            if not apigateway:
                typer.echo("Didn't find an existing API. Creating a new API")
                apigateway = apigateway_client.create_rest_api(
                    name=api_name,
                    description=service_meta.summary,
                    endpointConfiguration={
                        "types": ["REGIONAL"]
                    },
                    disableExecuteApiEndpoint=True
                )
            else:
                typer.echo("Found existing API.")

            typer.echo()
            progress.update(1)
            typer.echo()

            api_id: str = apigateway['id']

            root_resource_id: str = ""

            typer.echo()
            typer.echo("Nuking old resources.")
            typer.echo()

            # Clear out all resources (and by extension their methods)
            for resource in apigateway_client.get_resources(restApiId=api_id, limit=500)['items']:
                if resource['path'] == "/":
                    root_resource_id = resource['id']
                else:
                    try:
                        apigateway_client.delete_resource(
                            restApiId=api_id,
                            resourceId=resource['id']
                        )
                    except:
                        # Depending on the order that paths are retrieved in, we may delete a parent path before the child paths and that would cause an exception.
                        pass

            progress.update(2)
            typer.echo()

            typer.echo()
            typer.echo("Creating resources, methods, and integrations.")
            typer.echo()
            typer.echo("Reserved keywords: " + str(RESERVED_WORDS))
            typer.echo("Any endpoint with a reserved keyword inside of the path will be IGNORED.")
            typer.echo()
            typer.echo("Please wait...")
            typer.echo()

            # Maps a str path to the AWS API Gateway resource ID
            resources: Dict[str, str] = {}

            def create_resources(path_components: List[str]) -> str:
                """
                Recursive algorithm which builds up paths in API Gateway regardless of the order that the paths come in
                """
                full_path: str = "/".join(path_components)
                if full_path in resources:
                    return resources[full_path]

                if len(path_components) == 1:
                    parent_resource_id: str = root_resource_id
                else:
                    parent_resource_id: str = create_resources(path_components=path_components[0:-1])

                new_resource_id = apigateway_client.create_resource(
                    restApiId=api_id,
                    parentId=parent_resource_id,
                    pathPart=path_components[-1]
                )["id"]

                resources[full_path] = new_resource_id

                return new_resource_id

            for endpoint_path, endpoint_methods in api_spec["paths"].items():

                # Remove prefixed '/'
                full_unaltered_path: str = endpoint_path[1:]

                # Remove prefixed '/' and ignored prefixes
                endpoint_path = endpoint_path[1:]
                if ignore_prefixes > 0:
                    endpoint_path = endpoint_path.split("/")
                    endpoint_path = endpoint_path[ignore_prefixes:]
                    endpoint_path = "/".join(endpoint_path)

                # Ignore paths which contain a reserved word
                ignore: bool = False
                endpoint_path_components: List[str] = endpoint_path.split("/")
                for endpoint_path_component in endpoint_path_components:
                    if endpoint_path_component in RESERVED_WORDS:
                        ignore = True
                        break

                if ignore:
                    continue

                resource_id: str = create_resources(path_components=endpoint_path_components)

                methods: List[str] = ["OPTIONS"]

                # Adds methods from the spec: POST, GET, PUT, DELETE, etc.
                for endpoint_method, endpoint_details in endpoint_methods.items():
                    methods.append(endpoint_method.upper())

                    # Building the dicts for the method and integration, so that path parameters work
                    full_unaltered_path_components: List[str] = full_unaltered_path.split("/")
                    method_request_parameters: Dict[str, str] = {}
                    integration_request_parameters: Dict[str, str] = {}
                    for component in full_unaltered_path_components:
                        if component[0] == "{" and component[-1] == "}":
                            method_request_parameters[f"method.request.path.{component[1:-1]}"] = True
                            integration_request_parameters[
                                f"integration.request.path.{component[1:-1]}"] = f"method.request.path.{component[1:-1]}"

                    apigateway_client.put_method(
                        restApiId=api_id,
                        resourceId=resource_id,
                        httpMethod=endpoint_method.upper(),
                        authorizationType="NONE",
                        requestParameters=method_request_parameters
                    )

                    apigateway_client.put_integration(
                        restApiId=api_id,
                        resourceId=resource_id,
                        httpMethod=endpoint_method.upper(),
                        type="HTTP_PROXY",
                        integrationHttpMethod=endpoint_method.upper(),
                        uri=f"http://${{stageVariables.lb}}:{target_group_port}/{full_unaltered_path}",
                        requestParameters=integration_request_parameters,
                        connectionType="VPC_LINK",
                        connectionId="${stageVariables.vpc}"
                    )

                    apigateway_client.put_method_response(
                        restApiId=api_id,
                        resourceId=resource_id,
                        httpMethod=endpoint_method.upper(),
                        statusCode="200",
                        responseParameters={
                            'method.response.header.Access-Control-Allow-Origin': False
                        },
                        responseModels={
                            'application/json': 'Empty'
                        }
                    )

                # Sets up CORS by adding the OPTION method
                apigateway_client.put_method(
                    restApiId=api_id,
                    resourceId=resource_id,
                    httpMethod="OPTIONS",
                    authorizationType="NONE"
                )

                apigateway_client.put_integration(
                    restApiId=api_id,
                    resourceId=resource_id,
                    httpMethod="OPTIONS",
                    type="MOCK",
                    requestTemplates={
                        'application/json': '{"statusCode": 200}'
                    }
                )

                apigateway_client.put_method_response(
                    restApiId=api_id,
                    resourceId=resource_id,
                    httpMethod="OPTIONS",
                    statusCode="200",
                    responseParameters={
                        'method.response.header.Access-Control-Allow-Headers': False,
                        'method.response.header.Access-Control-Allow-Origin': False,
                        'method.response.header.Access-Control-Allow-Methods': False
                    },
                    responseModels={
                        'application/json': 'Empty'
                    }
                )

                method_str: str = ""
                for method in methods:
                    if len(method_str) == 0:
                        method_str += f"{method}"
                    else:
                        method_str += f",{method}"

                apigateway_client.put_integration_response(
                    restApiId=api_id,
                    resourceId=resource_id,
                    httpMethod="OPTIONS",
                    statusCode="200",
                    responseParameters={
                        'method.response.header.Access-Control-Allow-Headers': f"'{EHELPLY_HEADERS},Content-Type,X-Amz-Date,Authorization,X-Access-Token,X-Secret-Token,X-Api-Key,X-Amz-Security-Token'",
                        'method.response.header.Access-Control-Allow-Methods': f"'{method_str}'",
                        'method.response.header.Access-Control-Allow-Origin': "'*'"
                    },
                    responseTemplates={
                        'application/json': ''
                    }
                )

            progress.update(1)
            typer.echo()
            typer.echo()

            typer.echo("Checking for deployment stages")

            # Setting up stages if they do not already exist
            stages: List[str] = []
            for stage in apigateway_client.get_stages(restApiId=api_id)['item']:
                stages.append(stage['stageName'])

            created_test: bool = False

            from time import sleep

            if "test" not in stages:
                typer.echo()
                typer.echo("Creating a deployment stage for the 'test' environment")

                created_test = True

                apigateway_client.create_deployment(
                    restApiId=api_id,
                    stageName="test",
                    stageDescription="Test environment",
                    variables={'lb': 'eHelply-Test-Network-LB-8595beb95bd6a14a.elb.ca-central-1.amazonaws.com',
                               'vpc': 'hcp323'}
                )

                typer.echo()
                typer.echo("Adding WAF integration to the 'test' deployment stage")

                waf_client = State.aws.make_client(name="wafv2")

                waf_client.associate_web_acl(
                    WebACLArn=WEB_ACL_ARN,
                    ResourceArn=f"arn:aws:apigateway:ca-central-1::/restapis/{api_id}/stages/test"
                )

            if "prod" not in stages:
                typer.echo()
                typer.echo("Creating a deployment stage for the 'prod' environment")

                if created_test:
                    typer.echo("Waiting out API throttling... (15s)")
                    sleep(15)

                apigateway_client.create_deployment(
                    restApiId=api_id,
                    stageName="prod",
                    stageDescription="Production environment",
                    variables={'lb': 'eHelply-Prod-Network-LB-913c7576c63aee8d.elb.ca-central-1.amazonaws.com',
                               'vpc': 'pe40zx'}
                )

        typer.echo()
        progress.update(1)
        typer.echo()
        typer.echo()

        typer.echo("API Gateway sync complete.")
        typer.echo()
        typer.echo(
            "If this is your first time using API Gateway sync on this microservice, please remember to add or adjust custom domain name path mapping where needed.")
        typer.echo()
        typer.echo(
            "NOTE: This sync does not automatically deploy your API. This is a safety pre-caution. You will need to login to API Gateway and deploy to whichever deployment stages you desire.")

        sys.exit()




import traceback
import json
import base64
import time
from typing import Union, Optional, List, Any
from enum import Enum

import sentry_sdk
from sentry_sdk.tracing import Span
from fastapi import HTTPException
from pydantic import BaseModel, PrivateAttr

from ehelply_bootstrapper.utils.state import State
from ehelply_bootstrapper.drivers.aws_utils.aws_dynamo import Dynamo, Attribute
from ehelply_bootstrapper.utils.cryptography import Hashing

from ehelply_microservice_library.integrations.m2m import M2M

default_secondary: str = "s"


class AuthException(Exception):
    pass


class CloudParticipantAuthRequest(BaseModel):
    x_access_token: Optional[str] = None
    x_secret_token: Optional[str] = None
    authorization: Optional[str] = None
    ehelply_active_participant: Optional[str] = None
    ehelply_project: Optional[str] = None
    ehelply_data: Optional[str] = None


class CloudParticipantRequest(BaseModel):
    auth: CloudParticipantAuthRequest
    scopes: List[str]
    allow_project_statuses: List[str] = ["enabled"]
    limit: Optional[str] = None
    ignore_project_enabled: bool = False
    ignore_spend_limits: bool = False
    exception_if_unauthorized: bool = True
    exception_if_spend_maxed: bool = True
    exception_if_project_not_enabled: bool = True
    skip_project_check: bool = False


class CloudParticipantAuthResponse(BaseModel):
    authorization: Optional[str] = None
    access_token: Optional[str] = None
    secret_token: Optional[str] = None
    claims: Optional[dict] = None
    data: Optional[dict] = None


class CloudParticipantResponse(BaseModel):
    auth: CloudParticipantAuthResponse
    active_participant: str
    project_uuid: str
    role: str
    scopes: List[str]
    is_privileged: bool = False
    is_m2m: bool = False
    entity_identifier: Optional[str] = None
    limit: Optional[float] = None


class CloudParticipant(BaseModel):
    active_participant: str
    project_uuid: str
    is_privileged: bool = False
    entity_identifier: Optional[str] = None


class DynamoModel(BaseModel):
    __last_update: dict = PrivateAttr()

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self.__last_update: dict = self.dict()

    def get_pkey_names(self) -> List[str]:
        raise Exception("Requires implementation")

    def get_pkey_values(self) -> List[Any]:
        raise Exception("Requires implementation")

    def to_dynamo_dict(self) -> dict:
        raise Exception("Requires implementation")

    @staticmethod
    def dict_to_model(data: dict) -> 'DynamoModel':
        raise Exception("Requires implementation")

    def get_changed(self) -> List[Attribute]:
        attributes: List[Attribute] = []
        current: dict = self.dict()
        for key, value in current.items():
            if self.__last_update[key] != value:
                attributes.append(Attribute(
                    key=key,
                    value=value
                ))
        self.__last_update = current
        return attributes


class ProjectDB(DynamoModel):
    """
    Used for DB row entry
    """
    uuid: str
    name: str
    created_at: str
    current_spend: int  # Dollar formats represented by a x10000000 integer. Precision to the millonth
    max_spend: int  # Dollar formats represented by a x10000000 integer. Precision to the millonth
    project_status: str
    archived_at: Optional[str] = None

    def get_pkey_names(self) -> List[str]:
        return [
            "primary",
            "secondary"
        ]

    def get_pkey_values(self) -> List[Any]:
        return [
            self.uuid,
            default_secondary
        ]

    def to_dynamo_dict(self) -> dict:
        the_dict: dict = self.dict()
        the_dict["primary"] = self.uuid
        the_dict["secondary"] = default_secondary
        del the_dict["uuid"]
        return the_dict

    @staticmethod
    def dict_to_model(data: dict) -> 'ProjectDB':
        return ProjectDB(**data, uuid=data['primary'])

    def create(self):
        AuthDynamos.projects.create_item(self.get_pkey_values(), self.to_dynamo_dict())

    @staticmethod
    def get(uuid: str) -> 'ProjectDB':
        return ProjectDB.dict_to_model(
            AuthDynamos.projects.get_item([
                uuid,
                default_secondary
            ])
        )

    @staticmethod
    def delete(uuid: str):
        AuthDynamos.projects.delete_item([
            uuid,
            default_secondary
        ])

    def update(self, attributes: Optional[List[Attribute]] = None):
        if attributes is None:
            attributes = self.get_changed()
        AuthDynamos.projects.update_item(self.get_pkey_values(), attributes)

    @staticmethod
    def update_for(uuid: str, attributes: List[Attribute]):
        AuthDynamos.projects.update_item([
            uuid,
            default_secondary
        ], attributes)

    def is_spend_maxed(self) -> bool:
        return self.current_spend >= self.max_spend


class DefaultLimitsDB(DynamoModel):
    """
    Used for DB row entry
    """
    limit_name: str
    value: float

    def get_pkey_names(self) -> List[str]:
        return [
            "primary",
            "secondary"
        ]

    def get_pkey_values(self) -> List[Any]:
        return [
            self.limit_name,
            default_secondary
        ]

    def to_dynamo_dict(self) -> dict:
        the_dict: dict = self.dict()
        the_dict["primary"] = self.limit_name
        the_dict["secondary"] = default_secondary
        del the_dict["limit_name"]
        return the_dict

    @staticmethod
    def dict_to_model(data: dict) -> 'DefaultLimitsDB':
        return DefaultLimitsDB(**data, limit_name=data['primary'])

    def create(self):
        AuthDynamos.default_limits.create_item(self.get_pkey_values(), self.to_dynamo_dict())

    @staticmethod
    def get(default_limit: str) -> 'DefaultLimitsDB':
        return DefaultLimitsDB.dict_to_model(
            AuthDynamos.default_limits.get_item([
                default_limit,
                default_secondary
            ])
        )

    @staticmethod
    def delete(default_limit: str):
        AuthDynamos.default_limits.delete_item([
            default_limit,
            default_secondary
        ])

    def update(self, attributes: Optional[List[Attribute]] = None):
        if attributes is None:
            attributes = self.get_changed()
        AuthDynamos.default_limits.update_item(self.get_pkey_values(), attributes)

    @staticmethod
    def update_for(default_limit: str, attributes: List[Attribute]):
        AuthDynamos.default_limits.update_item([
            default_limit,
            default_secondary
        ], attributes)


class ProjectMemberRoleEnum(str, Enum):
    owner = 'owner'
    member = 'member'
    key = 'key'


class ProjectMembersDB(DynamoModel):
    """
    Used for DB row entry
    """
    project_uuid: str
    member_uuid: str
    role: ProjectMemberRoleEnum
    scopes: List[str]

    def get_pkey_names(self) -> List[str]:
        return [
            "primary",
            "secondary"
        ]

    def get_pkey_values(self) -> List[Any]:
        return [
            self.project_uuid,
            self.member_uuid
        ]

    def to_dynamo_dict(self) -> dict:
        the_dict: dict = self.dict()
        the_dict["primary"] = self.project_uuid
        the_dict["secondary"] = self.member_uuid
        del the_dict["project_uuid"]
        del the_dict["member_uuid"]
        return the_dict

    @staticmethod
    def dict_to_model(data: dict) -> 'ProjectMembersDB':
        return ProjectMembersDB(**data, project_uuid=data['primary'], member_uuid=data['secondary'])

    def create(self):
        AuthDynamos.project_members.create_item(self.get_pkey_values(), self.to_dynamo_dict())

    @staticmethod
    def get(project_uuid: str, member_uuid: str) -> 'ProjectMembersDB':
        return ProjectMembersDB.dict_to_model(
            AuthDynamos.project_members.get_item([
                project_uuid,
                member_uuid
            ])
        )

    @staticmethod
    def delete(project_uuid: str, member_uuid: str):
        AuthDynamos.project_members.delete_item([
            project_uuid,
            member_uuid
        ])

    def update(self, attributes: Optional[List[Attribute]] = None):
        if attributes is None:
            attributes = self.get_changed()
        AuthDynamos.project_members.update_item(self.get_pkey_values(), attributes)

    @staticmethod
    def update_for(project_uuid: str, member_uuid: str, attributes: List[Attribute]):
        AuthDynamos.project_members.update_item([
            project_uuid,
            member_uuid
        ], attributes)


class ProjectLimitsDB(DynamoModel):
    """
    Used for DB row entry
    """
    project_uuid: str
    limit_name: str
    value: float

    def get_pkey_names(self) -> List[str]:
        return [
            "primary",
            "secondary"
        ]

    def get_pkey_values(self) -> List[Any]:
        return [
            self.project_uuid,
            self.limit_name
        ]

    def to_dynamo_dict(self) -> dict:
        the_dict: dict = self.dict()
        the_dict["primary"] = self.project_uuid
        the_dict["secondary"] = self.limit_name
        del the_dict["project_uuid"]
        del the_dict["limit_name"]
        return the_dict

    @staticmethod
    def dict_to_model(data: dict) -> 'ProjectLimitsDB':
        return ProjectLimitsDB(**data, project_uuid=data['primary'], limit_name=data['secondary'])

    def create(self):
        AuthDynamos.project_limits.create_item(self.get_pkey_values(), self.to_dynamo_dict())

    @staticmethod
    def get(project_uuid: str, limit_name: str) -> 'ProjectLimitsDB':
        return ProjectLimitsDB.dict_to_model(
            AuthDynamos.project_limits.get_item([
                project_uuid,
                limit_name
            ])
        )

    @staticmethod
    def delete(project_uuid: str, limit_name: str):
        AuthDynamos.project_limits.delete_item([
            project_uuid,
            limit_name
        ])

    def update(self, attributes: Optional[List[Attribute]] = None):
        if attributes is None:
            attributes = self.get_changed()
        AuthDynamos.project_limits.update_item(self.get_pkey_values(), attributes)

    @staticmethod
    def update_for(project_uuid: str, limit_name: str, attributes: List[Attribute]):
        AuthDynamos.project_limits.update_item([
            project_uuid,
            limit_name
        ], attributes)


class StackAdministratorsDB(DynamoModel):
    """
    Used for DB row entry
    """
    member_uuid: str
    scopes: List[str]

    def get_pkey_names(self) -> List[str]:
        return [
            "primary",
            "secondary"
        ]

    def get_pkey_values(self) -> List[Any]:
        return [
            self.member_uuid,
            default_secondary
        ]

    def to_dynamo_dict(self) -> dict:
        the_dict: dict = self.dict()
        the_dict["primary"] = self.member_uuid
        the_dict["secondary"] = default_secondary
        del the_dict["member_uuid"]
        return the_dict

    @staticmethod
    def dict_to_model(data: dict) -> 'StackAdministratorsDB':
        return StackAdministratorsDB(**data, member_uuid=data['primary'])

    def create(self):
        AuthDynamos.stack_administrators.create_item(self.get_pkey_values(), self.to_dynamo_dict())

    @staticmethod
    def get(member_uuid: str) -> 'StackAdministratorsDB':
        return StackAdministratorsDB.dict_to_model(
            AuthDynamos.stack_administrators.get_item([
                member_uuid,
                default_secondary
            ])
        )

    @staticmethod
    def delete(member_uuid: str):
        AuthDynamos.stack_administrators.delete_item([
            member_uuid,
            default_secondary
        ])

    def update(self, attributes: Optional[List[Attribute]] = None):
        if attributes is None:
            attributes = self.get_changed()
        AuthDynamos.stack_administrators.update_item(self.get_pkey_values(), attributes)

    @staticmethod
    def update_for(member_uuid: str, attributes: List[Attribute]):
        AuthDynamos.stack_administrators.update_item([
            member_uuid,
            default_secondary
        ], attributes)


class KeysDB(DynamoModel):
    """
    Used for DB row entry
    """
    access_token: str
    secret_token: str

    def get_pkey_names(self) -> List[str]:
        return [
            "primary",
            "secondary"
        ]

    def get_pkey_values(self) -> List[Any]:
        return [
            self.access_token,
            default_secondary
        ]

    def to_dynamo_dict(self) -> dict:
        the_dict: dict = self.dict()
        the_dict["primary"] = self.access_token
        the_dict["secondary"] = default_secondary
        del the_dict["access_token"]
        return the_dict

    @staticmethod
    def dict_to_model(data: dict) -> 'KeysDB':
        return KeysDB(**data, access_token=data['primary'])

    def create(self):
        AuthDynamos.keys.create_item(self.get_pkey_values(), self.to_dynamo_dict())

    @staticmethod
    def get(access_token: str) -> 'KeysDB':
        return KeysDB.dict_to_model(
            AuthDynamos.keys.get_item([
                access_token,
                default_secondary
            ])
        )

    @staticmethod
    def delete(access_token: str):
        AuthDynamos.keys.delete_item([
            access_token,
            default_secondary
        ])

    def update(self, attributes: Optional[List[Attribute]] = None):
        if attributes is None:
            attributes = self.get_changed()
        AuthDynamos.keys.update_item(self.get_pkey_values(), attributes)

    @staticmethod
    def update_for(access_token: str, attributes: List[Attribute]):
        AuthDynamos.keys.update_item([
            access_token,
            default_secondary
        ], attributes)


class AuthDynamos:
    default_limits: Dynamo
    projects: Dynamo
    project_members: Dynamo
    project_limits: Dynamo
    stack_administrators: Dynamo
    keys: Dynamo
    loaded: bool = False

    @staticmethod
    def make():
        if not AuthDynamos.loaded:
            table_name: str = State.config.app.auth.dynamo_table

            AuthDynamos.default_limits = Dynamo(
                table_name=table_name,
                p_keys=["primary", "secondary"],
                sub_table="default_limits"
            )
            AuthDynamos.projects = Dynamo(
                table_name=table_name,
                p_keys=["primary", "secondary"],
                sub_table="projects"
            )
            AuthDynamos.project_members = Dynamo(
                table_name=table_name,
                p_keys=["primary", "secondary"],
                sub_table="project_members"
            )
            AuthDynamos.project_limits = Dynamo(
                table_name=table_name,
                p_keys=["primary", "secondary"],
                sub_table="project_limits"
            )
            AuthDynamos.stack_administrators = Dynamo(
                table_name=table_name,
                p_keys=["primary", "secondary"],
                sub_table="stack_administrators"
            )
            AuthDynamos.keys = Dynamo(
                table_name=table_name,
                p_keys=["primary", "secondary"],
                sub_table="keys"
            )
            AuthDynamos.loaded = True


async def ehelply_cloud_access(
        auth: CloudParticipantAuthRequest,
        scopes: List[str],
        allow_project_statuses: Optional[List[str]] = None,
        limit: Optional[str] = None,
        ignore_project_enabled=False,
        ignore_spend_limits=False,
        exception_if_unauthorized=True,
        exception_if_spend_maxed=True,
        exception_if_project_not_enabled=True,
        skip_project_check=False,
) -> Union[bool, CloudParticipantResponse]:
    """

    Args:
        limit:
        allow_project_statuses:
        auth:
        node:
        service_target:
        ignore_project_enabled:
        ignore_spend_limits:
        exception_if_unauthorized:
        exception_if_spend_maxed:
        exception_if_project_not_enabled:
        skip_project_check:

    Returns:

    """
    cloud_participant_request: CloudParticipantRequest = CloudParticipantRequest(
        auth=auth,
        scopes=scopes,
        allow_project_statuses=allow_project_statuses if allow_project_statuses is not None else ["enabled"],
        limit=limit,
        ignore_project_enabled=ignore_project_enabled,
        ignore_spend_limits=ignore_spend_limits,
        exception_if_unauthorized=exception_if_unauthorized,
        exception_if_spend_maxed=exception_if_spend_maxed,
        exception_if_project_not_enabled=exception_if_project_not_enabled,
        skip_project_check=skip_project_check
    )

    start_time = time.time()
    AuthDynamos.make()
    response: Union[bool, CloudParticipantResponse] = __cloud_participant(cloud_participant_request)
    State.logger.debug("CloudParticipant Auth: --- %s seconds ---" % (time.time() - start_time))
    State.logger.debugg(str(response.dict()))
    return response


def __cloud_participant(cloud_participant_request: CloudParticipantRequest) -> Union[bool, CloudParticipantResponse]:
    with sentry_sdk.start_span(
            op="cp_authentication",
            description="Cloud Participant Authentication"
    ) as sentry_span:
        m2m: M2M = State.integrations.get("m2m")
        ehelply_data_dict, ehelply_entity_identifier, ehelply_active_participant, ehelply_project_uuid, my_auth, is_m2m_key = __cp_authentication(
            cloud_participant_request, m2m, sentry_span
        )

    with sentry_sdk.start_span(
            op="cp_authorization",
            description="Cloud Participant Authorization"
    ) as sentry_span:
        if is_m2m_key:
            response: CloudParticipantResponse = CloudParticipantResponse(
                auth=my_auth,
                is_privileged=True,
                is_m2m_key=True,
                active_participant=ehelply_active_participant,
                project_uuid=ehelply_project_uuid,
                entity_identifier=ehelply_entity_identifier,
                role="stack_administrator",
                scopes=[],
                limit=None
            )
            sentry_span.set_tag('is_speedy_m2m', True)
            sentry_span.set_data('CloudParticipantResponse', __remove_sensitive_data(response, safe_participant=True))
            return response

        try:
            with sentry_span.start_child(op="cp_authorization", description="Retrieving Project Details"):
                project: ProjectDB = ProjectDB.get(ehelply_project_uuid)
        except:
            raise HTTPException(
                status_code=401,
                detail="Invalid project - Denied by eHelply - CloudParticipant Auth"
            )

        role: str
        scopes: List[str]
        is_privileged: bool = False
        limit: Optional[float] = None

        try:
            with sentry_span.start_child(op="cp_authorization", description="Retrieving Project Member"):
                project_member: ProjectMembersDB = ProjectMembersDB.get(ehelply_project_uuid,
                                                                        ehelply_active_participant)

                for scope in cloud_participant_request.scopes:
                    if scope not in project_member.scopes:
                        raise HTTPException(
                            status_code=401,
                            detail="Unauthorized - Denied by eHelply - CloudParticipant Auth"
                        )

                role = project_member.role
                scopes = project_member.scopes

        except:
            try:
                with sentry_span.start_child(op="cp_authorization", description="Retrieving Stack Administrator"):
                    stack_administrator: StackAdministratorsDB = StackAdministratorsDB.get(ehelply_active_participant)

                    for scope in cloud_participant_request.scopes:
                        if scope not in stack_administrator.scopes:
                            raise HTTPException(
                                status_code=401,
                                detail="Unauthorized - Denied by eHelply - CloudParticipant Auth"
                            )

                    role = "stack_administrator"
                    scopes = stack_administrator.scopes
                    is_privileged = True
            except:
                raise HTTPException(
                    status_code=401,
                    detail="Invalid participant/key - Denied by eHelply - CloudParticipant Auth"
                )

        if cloud_participant_request.limit:
            try:
                with sentry_span.start_child(op="cp_authorization", description="Retrieving Project Limit"):
                    project_limits: ProjectLimitsDB = ProjectLimitsDB.get(ehelply_project_uuid,
                                                                          cloud_participant_request.limit)

                    limit = project_limits.value
            except:
                raise HTTPException(
                    status_code=500,
                    detail="Invalid limit - Error by eHelply - CloudParticipant Auth"
                )

        spend_maxed: bool = False
        project_not_enabled: bool = False

        try:
            if not cloud_participant_request.ignore_project_enabled \
                    and project.project_status not in cloud_participant_request.allow_project_statuses:
                project_not_enabled = True
                raise AuthException("Project is not enabled")

            if not cloud_participant_request.ignore_spend_limits and project.is_spend_maxed():
                spend_maxed = True
                raise AuthException("Spend is maxed")

            response: CloudParticipantResponse = CloudParticipantResponse(
                auth=my_auth,
                is_privileged=is_privileged,
                active_participant=ehelply_active_participant,
                project_uuid=ehelply_project_uuid,
                entity_identifier=ehelply_entity_identifier,
                role=role,
                scopes=scopes,
                limit=limit
            )
            sentry_span.set_tag('is_speedy_m2m', False)
            sentry_span.set_data('CloudParticipantResponse', __remove_sensitive_data(response))
            return response
        except (AuthException, HTTPException):
            pass
        except Exception as e:
            traceback.print_exc()
            State.logger.warning(message="Project checks caused an unexpected error.")
            raise HTTPException(
                status_code=500,
                detail="Ruh Roh, something has gone terribly wrong - Error by eHelply - CloudParticipant Auth"
            )

        if spend_maxed and cloud_participant_request.exception_if_spend_maxed:
            raise HTTPException(
                status_code=400,
                detail="Project spend is maxed - Denied by eHelply - CloudParticipant Auth. Increase max spend to complete this request."
            )
        elif spend_maxed:
            return False

        if project_not_enabled and cloud_participant_request.exception_if_project_not_enabled:
            raise HTTPException(
                status_code=400,
                detail="Project is not enabled - Denied by eHelply - CloudParticipant Auth. The project may be awaiting approval or disabled."
            )
        elif project_not_enabled:
            return False

        if cloud_participant_request.exception_if_unauthorized:
            raise HTTPException(
                status_code=403,
                detail="Unauthorized - Denied by eHelply - CloudParticipant Auth"
            )
        else:
            return False


def __cp_authentication(cloud_participant_request: CloudParticipantRequest, m2m: M2M, sentry_span: Span):
    ehelply_active_participant: Optional[str] = None

    with sentry_span.start_child(op="cp_authentication", description="Parsing eHelply Data Header"):
        ehelply_data_dict, ehelply_entity_identifier = __parse_ehelply_data(cloud_participant_request)

    my_auth = CloudParticipantAuthResponse(
        data=ehelply_data_dict
    )

    if cloud_participant_request.auth.authorization:
        with sentry_span.start_child(op="cp_authentication", description="Authorization Token Authentication"):
            ehelply_active_participant = __authentication_by_authorization_token(
                cloud_participant_request,
                m2m,
                my_auth
            )

    is_m2m_key: bool = False

    if cloud_participant_request.auth.x_secret_token and cloud_participant_request.auth.x_access_token:
        with sentry_span.start_child(op="cp_authentication", description="Key (Access/Secret) Authentication"):
            ehelply_active_participant, is_m2m_key = __authentication_by_key(
                cloud_participant_request,
                my_auth,
                m2m
            )

    ehelply_project_uuid = cloud_participant_request.auth.ehelply_project

    if not ehelply_active_participant:
        raise HTTPException(
            status_code=401,
            detail="Missing participant - Denied by eHelply - CloudParticipant Auth"
        )

    if not ehelply_project_uuid:
        raise HTTPException(
            status_code=401,
            detail="Missing project - Denied by eHelply - CloudParticipant Auth"
        )

    return ehelply_data_dict, ehelply_entity_identifier, ehelply_active_participant, ehelply_project_uuid, my_auth, is_m2m_key


def __authentication_by_key(
        cloud_participant_request: CloudParticipantRequest,
        my_auth: CloudParticipantAuthResponse,
        m2m: M2M
) -> str:
    my_auth.access_token = cloud_participant_request.auth.x_access_token
    my_auth.secret_token = cloud_participant_request.auth.x_secret_token

    try:
        is_m2m_key: bool = False
        is_valid_key: bool = m2m.speedy_m2m_auth(
            cloud_participant_request.auth.x_access_token,
            cloud_participant_request.auth.x_secret_token
        )
        if not is_valid_key:
            is_valid_key = __verify_key(
                access=cloud_participant_request.auth.x_access_token,
                secret=cloud_participant_request.auth.x_secret_token
            )
        else:
            is_m2m_key = True
    except:
        raise HTTPException(
            status_code=401,
            detail="Invalid key - Denied by eHelply - CloudParticipant Auth"
        )

    if not is_valid_key:
        raise HTTPException(
            status_code=401,
            detail="Invalid key - Denied by eHelply - CloudParticipant Auth"
        )

    ehelply_active_participant = my_auth.access_token

    return ehelply_active_participant, is_m2m_key


def __authentication_by_authorization_token(
        cloud_participant_request: CloudParticipantRequest,
        m2m: M2M,
        my_auth: CloudParticipantAuthResponse
) -> str:
    ehelply_active_participant: Optional[str] = None
    try:
        claims = m2m.verify_token(token=cloud_participant_request.auth.authorization)
    except:
        raise HTTPException(
            status_code=401,
            detail="Invalid token - Denied by eHelply - CloudParticipant Auth"
        )
    my_auth.authorization = cloud_participant_request.auth.authorization
    if cloud_participant_request.auth.ehelply_active_participant:
        if cloud_participant_request.auth.ehelply_active_participant in claims["custom:participants"].split(","):
            ehelply_active_participant = cloud_participant_request.auth.ehelply_active_participant
        else:
            raise HTTPException(
                status_code=401,
                detail="Invalid participant - Denied by eHelply - CloudParticipant Auth"
            )
    my_auth.claims = claims
    return ehelply_active_participant


def __parse_ehelply_data(
        cloud_participant_request: CloudParticipantRequest
) -> (dict, str):
    ehelply_data_dict: Optional[dict] = None
    ehelply_entity_identifier: Optional[str] = None

    if cloud_participant_request.auth.ehelply_data:
        try:
            ehelply_data_b64_bytes = cloud_participant_request.auth.ehelply_data.encode("utf-8")
            ehelply_data_bytes = base64.b64decode(ehelply_data_b64_bytes)
            ehelply_data_dict = json.loads(ehelply_data_bytes.decode("utf-8"))

            if "entity_identifier" in ehelply_data_dict:
                ehelply_entity_identifier = ehelply_data_dict["entity_identifier"]

        except:
            raise HTTPException(
                status_code=400,
                detail="Invalid data - Denied by eHelply - CloudParticipant Auth"
            )
    return ehelply_data_dict, ehelply_entity_identifier


def __verify_key(access: str, secret: str) -> bool:
    actual_secret: str = KeysDB.get(access).secret_token

    hasher: Hashing = Hashing()

    if not hasher.check(secret, actual_secret):
        return False

    return True


def __remove_sensitive_data(cloud_participant_response: CloudParticipantResponse, safe_participant: bool = False) -> dict:
    response: dict = cloud_participant_response.dict()
    if response['auth']['secret_token']:
        response['auth']['secret_token'] = "***"

    if response['role'] == 'key' or safe_participant and response['active_participant'].length > 8:
        response['active_participant'] = f"...{response['active_participant'][-8:]}"

    if response['auth']['access_token'] and response['auth']['access_token'].length > 8:
        response['auth']['access_token'] = f"...{response['auth']['access_token'][-8:]}"

    return response



"""
3rd Party Libraries
Not much interesting here. We have some default imports for FastAPI endpoints and common SQL Alchemy use cases
"""
from fastapi import APIRouter, Body

from ehelply_microservice_library.integrations.fact import get_fact_endpoint
from ehelply_microservice_library.integrations.sdk import SDK

"""
eHelply Bootstrapper
Now we are getting interesting.
- State provides you with access to the app state
- get_db is used with Dependency injection to obtain a fresh DB connection session
    `db: Session = Depends(get_db)`
- Responses imports a bunch of default HTTP responses that eHelply has standardized on. Use these where possible as your
    return from each endpoint
"""
from ehelply_bootstrapper.utils.state import State

from ehelply_bootstrapper.drivers.fast_api_utils.responses import *

from ehelply_batcher.abstract_timer_service import AbstractTimerService
from ehelply_logger.Logger import Logger
import threading
import asyncio
import requests

"""
Router
This is a self-contained router instance. By itself it does nothing, but by attaching endpoints to it using decorator
  notation, you can make an epic router!
"""
router = APIRouter()


@router.post(
    '/keys/categories/{category_name}',
    tags=["keys"],
)
async def add_key(category_name: str, data: dict = Body({})):
    """
    Add a new key
    :return:
    """
    State.secrets.add(data['key'], category=category_name)

    return http_200_ok({"message": "Added key"})


@router.delete(
    '/keys/categories/{category_name}',
    tags=["keys"],
)
async def remove_key(category_name: str, data: dict = Body({})):
    """
    Add a new key
    :return:
    """
    State.secrets.remove(data['key'], category=category_name)

    return http_200_ok({"message": "Removed key"})


def get_encryption_keys(category: str, secret_key: str):
    return requests.get(
        get_fact_endpoint("ehelply-security") + f"/encryption/categories/{category}/keys",
        headers={'EHELPLY-SECURITY-SECRET-KEY': secret_key}
    ).json()


def create_encryption_key(category: str, secret_key: str):
    return requests.post(
        get_fact_endpoint("ehelply-security") + f"/encryption/categories/{category}/keys",
        headers={'EHELPLY-SECURITY-SECRET-KEY': secret_key}
    ).json()


async def refresh_encryption_keys(category: str):
    from ehelply_bootstrapper.utils.db_encryption import DBEncryption
    from ehelply_microservice_library.utils.secret_names import secret_name_security_vault

    secret_key: str = State.secrets.get(secret_name_security_vault())[0].secret_key

    keys: list = get_encryption_keys(category=category, secret_key=secret_key)

    if not keys:
        create_encryption_key(category=category, secret_key=secret_key)
        keys = get_encryption_keys(category=category, secret_key=secret_key)

    for key in keys:
        if not State.secrets.exists(secret=key['key'], category='database'):
            State.secrets.add(key['key'], category='database', prepend=True)
            DBEncryption.update_keys()


class SecurityKeys(AbstractTimerService):
    """
    Periodically checks for new or removed keys
    """

    def __init__(self, logger: Logger, delay_seconds: int, category: str):
        self.category: str = category

        super().__init__(
            name="SecurityKeys",
            delay_seconds=delay_seconds,
            logger=logger,
        )

    async def proc(self):
        await refresh_encryption_keys(category=self.category)


class SecurityKeysThread(threading.Thread):
    def __init__(self, logger: Logger, delay: int, category: str):
        super().__init__()
        self.timer: SecurityKeys = None
        self.logger = logger
        self.delay: int = delay
        self.category: str = category
        self.daemon = True

    def run(self) -> None:
        self.timer = SecurityKeys(
            logger=self.logger,
            delay_seconds=self.delay,
            category=self.category
        )
        asyncio.run(self.timer.run())



from typing import List, Union, Tuple
from pathlib import Path
import glob
import os
import json
import asyncio

from slugify import slugify

from ehelply_bootstrapper.bootstrap import Bootstrap, LOADABLE_MYSQL
from ehelply_bootstrapper.drivers.mysql import MySQLCredentials
from ehelply_bootstrapper.drivers.aws_utils.aws_asm import ASM
from ehelply_bootstrapper.utils.state import State
from ehelply_bootstrapper.utils.environment import Environment
from ehelply_bootstrapper.utils.service import ServiceMeta, ServiceStatus, KPI
from ehelply_bootstrapper.utils.cryptography import Hashing

from ehelply_logger.Logger import VERBOSITY_NORMAL

from ehelply_microservice_library.integrations.m2m import M2M
from ehelply_microservice_library.integrations.sdk import SDK
from ehelply_microservice_library.integrations.fact import Fact, FactBase, get_fact_endpoint, get_fact_stage
from ehelply_microservice_library.integrations.logging import Log
from ehelply_microservice_library.integrations.monitor import Monitor
from ehelply_microservice_library.integrations.notifications import Notifications, NotificationType

from ehelply_microservice_library.utils.constants.constants_gatekeeper import *
from ehelply_microservice_library.utils.service_updater import ServiceUpdater
from ehelply_microservice_library.utils.secret_names import secret_name_database, secret_name_facts_vault, \
    secret_name_security_vault, secret_name_updater_vault

from ehelply_microservice_library.utils.ehelply_cloud import AuthDynamos

CONST_INTEGRATION_ACCESS: str = "access"
CONST_INTEGRATION_APPOINTMENTS: str = "appointments"
CONST_INTEGRATION_META: str = "meta"
CONST_INTEGRATION_NOTES: str = "notes"
CONST_INTEGRATION_NOTIFICATIONS: str = "notifications"
CONST_INTEGRATION_PLACES: str = "places"
CONST_INTEGRATION_REVIEWS: str = "reviews"
CONST_INTEGRATION_SECURITY: str = "security"
CONST_INTEGRATION_USERS: str = "users"
CONST_INTEGRATION_PRODUCTS: str = "products"


class ServiceBootstrap(Bootstrap):
    """
    This class helps to set up the entire service

    In the default configuration, some assumptions are made about surrounding infrastructure. However, if these
    assumptions are wrong and require changing, this class is built to be as flexible as possible.

    Assumptions:
        * A facts microservice is already deployed. This is a service that shares constants with all services and do not require security
        * A monitoring microservice is already deployed. This will take in heartbeats and help manage exceptions
        * A security microservice is already deployed. This handles tokens, access keys, and encryption keys
        * An updater microservice is already deployed. This handles automatic updating of microservice template code
        * AWS ASM is being utilized to hold secrets and constants which require security

    """

    def __init__(self):
        """
        Prepares the bootstrapper class.

        Ignore this method unless you need to init any values for the microservice itself.
        """
        service_environment_path: str = str(self.get_service_environment_path())
        Environment(path=service_environment_path)

        service_gatekeeper_key: str = self.get_service_gatekeeper_key()

        service_loadables: List[str] = self.get_service_loadables()
        service_configs: Union[List[str], None] = self.get_service_configs()
        service_verbosity: int = self.get_service_verbosity()
        service_config_path: str = str(self.get_service_config_path())

        service_name: str = self.get_service_name()
        service_key: str = slugify(service_name)
        service_version: str = self.get_service_version()
        service_summary: str = self.get_service_summary()
        service_authors: List[str] = self.get_service_authors()
        service_author_emails: List[str] = self.get_service_author_emails()

        service_meta: ServiceMeta = ServiceMeta(
            name=service_name,
            key=service_key,
            version=service_version,
            summary=service_summary,
            authors=service_authors,
            author_emails=service_author_emails
        )

        self.service_status: ServiceStatus = ServiceStatus()

        super().__init__(
            service_meta=service_meta,
            service_gatekeeper_key=service_gatekeeper_key,
            service_environment_path=service_environment_path,
            service_loadables=service_loadables,
            service_config_path=service_config_path,
            service_configs=service_configs,
            service_verbosity=service_verbosity
        )

    def get_service_name(self) -> str:
        """
        Returns the service name

        Returns:
            str: Service name

        """
        raise Exception(
            "get_service_name must be overridden in your service file. Please see `service.py`")

    def get_service_version(self) -> str:
        """
        Returns the service version

        Returns:
            str: Service version in semver format. eg 1.3.13

        """
        raise Exception(
            "get_service_version must be overridden in your service file. Please see `service.py`")

    def get_service_loadables(self) -> List[str]:
        """
        This defines which drivers should be loaded when the service runs
        Each driver has an associated constant
        Add the applicable constant to the list for the drivers you wish to use

        Returns:
            List[str]: A list of loadables. These are the drivers that will be loaded when the service boots

        """
        raise Exception(
            "get_service_loadables must be overridden in your service file. Please see `service.py`")

    def get_service_required_integrations(self) -> List[str]:
        """
        This defines which integrations are required.

        Returns:
            List[str]: A list of integrations

        """
        raise Exception(
            "get_service_required_integrations must be overridden in your service file. Please see `service.py`")

    def get_service_optional_integrations(self) -> List[str]:
        """
        This defines which integrations are optional.

        Returns:
            List[str]: A list of integrations

        """
        return []

    def get_service_summary(self) -> str:
        """
        A summary of the service

        Returns:
            str: Service summary

        """
        raise Exception(
            "get_service_summary must be overridden in your service file. Please see `service.py`")

    def get_service_authors(self) -> List[str]:
        """
        Service authors

        Returns:
            List[str]: A list of author names

        """
        raise Exception(
            "get_service_authors must be overridden in your service file. Please see `service.py`")

    def get_service_author_emails(self) -> List[str]:
        """
        Email addresses of the service authors

        Returns:
            List[str]: A list of author emails in the same order as get_service_authors

        """
        raise Exception(
            "get_service_author_emails must be overridden in your service file. Please see `service.py`")

    def get_service_gatekeeper_key(self) -> str:
        """
        Gets the gatekeeper key from the environment variables. The gatekeeper key is an encryption key that decrypts
        any initial credentials which are required to start or use a microservice.

        Returns:
            str: Gatekeeper key

        """
        return json.loads(os.environ[GATEKEEPER_KEY_ENVIRONMENT_NAME])['secret_key']

    def get_root_path(self) -> Path:
        """
        Gets the root path

        Returns:
            Path: The root path

        """
        raise Exception("get_root_path must be overridden in service.py to be used.")

    def get_https_certificate_paths(self) -> Tuple[Union[None, str], Union[None, str]]:
        """
        Gets the certificates to run the dev server as https

        If, we return None, None the server will be launched as http

        Returns:
            Tuple[Union[None, str], Union[None, str]]: HTTPS certificate paths. First path is private key, second is cert

        """
        # certificates_path: Path = Path(Path(__file__).resolve().parents[2]).joinpath('certificates')
        #
        # return str(certificates_path.joinpath('private.key')), str(certificates_path.joinpath('certificate.crt'))
        return None, None

    def get_service_configs(self) -> Union[List[str], None]:
        """
        This is used to define the names (including .yaml) of other configuration files you would like to have the application
        load for you. If you are satisfied with the app.yaml and bootstrap.yaml configuration files, leave this value at None

        Returns:
            Union[List[str], None]: Additional

        """
        configs: List[str] = []

        service_configs_path: Path = self.get_service_config_path()

        for config in glob.glob(str(service_configs_path.joinpath('service')) + "/*.yaml"):
            config: Path = Path(config)
            configs.append(str(config.relative_to(service_configs_path)))

        return configs

    def get_service_verbosity(self) -> int:
        """

        This is used to define the level of debug we want to preform. Typically this is just used within the logger to
        output different amounts of information. The typical range is between 0 (No debug) to 3 (Full Debug)

        Returns:
            None

        """
        return VERBOSITY_NORMAL

    def get_service_environment_path(self) -> str:
        """
        This is an override for the environment path

        Returns:
            str: Path of the env.json file

        """
        return None

    def get_service_config_path(self) -> Path:
        """
        This is an override for the configuration path

        Returns:
            str: Path of configs folder

        """
        return None

    def get_service_package_path(self) -> Path:
        """
        This is an override for the microservice path

        Returns:
            str: Path of the microservice root

        """
        return None

    def get_fact_subscriptions(self) -> List[str]:
        """
        List of facts to subscribe to

        Returns:
            List[str]: List of facts to subscribe to and pull down from a facts microservice.

        """
        return []

    def get_fact_providers(self) -> List[FactBase]:
        """
        List of facts to push at a regular basis

        Returns:
            List[FactBase]: A list of facts to push to a facts microservice

        """
        return []

    def get_notification_types(self) -> List[NotificationType]:
        """
        List of notification types to push when the microservice loads

        Returns:
            List[NotificationType]: Notification types to be used throughout this service

        """
        return []

    def get_kpi(self) -> List[KPI]:
        """
        List of up to date KPIs

        Returns:
            List[KPI]: KPIs to report

        """
        return []

    def is_load_service_routers(self) -> bool:
        """
        Whether or not to register the service routers or not
        Returns:
            Bool
        """
        return True

    def fastapi_routers(self):
        """
        Register fastapi routers

        Returns:
            None

        """

        # !!!!! Imports for routers must go here to remove an order of operation bug that occurs when endpoints need to
        #          talk to a DB

        if self.is_load_service_routers():
            from ehelply_microservice_library.routers import router_service_monitor
            self.fastapi_driver.instance.include_router(
                router_service_monitor.router,
                prefix="/ehelply_monitor",
                tags=["ehelply", "service"],
                responses={404: {"description": "Not found"}},
            )

            from ehelply_microservice_library.routers import router_service_meta as service_meta_router
            self.fastapi_driver.instance.include_router(
                service_meta_router.router,
                prefix="/ehelply_meta",
                tags=["ehelply", "service"],
                responses={404: {"description": "Not found"}},
            )

            from ehelply_microservice_library.routers import router_service as service_router
            self.fastapi_driver.instance.include_router(
                service_router.router,
                prefix="/ehelply_service",
                tags=["ehelply", "service"],
                responses={404: {"description": "Not found"}},
            )

            from ehelply_microservice_library.routers import router_service_security as service_security_router
            self.fastapi_driver.instance.include_router(
                service_security_router.router,
                prefix="/ehelply_security",
                tags=["ehelply", "service"],
                responses={404: {"description": "Not found"}},
            )

    def get_secret_subscriptions(self) -> List[str]:
        """
        Hook into adding additional secret names to fetch during microservice boot

        Returns:
            List[str]: A list of secrets to subscribe to

        """
        return []

    def fetch_secrets(self):
        """
        Fetches secrets from AWS ASM

        Returns:
            None

        """
        subscribe_to_secrets: list = [
            secret_name_facts_vault(),
            secret_name_security_vault(),
            secret_name_updater_vault()
        ]

        if LOADABLE_MYSQL in self.get_service_loadables():
            subscribe_to_secrets.append(secret_name_database())

        subscribe_to_secrets += self.get_secret_subscriptions()

        for secret_name in subscribe_to_secrets:
            State.secrets.add(ASM.get_secret(secret_name=secret_name), category=secret_name)

    def register_integrations(self):
        """
        Register service integrations
        A service integration is a wrapper around ehelply microservices to make them easier to talk to

        Returns:
            None

        """
        # The fact service integration works on a subscription model. You provide the integration with a list of facts
        #   you care about and it will grab the facts from the fact service for you and ensure it stays updated for
        #   if/when the value of the fact changes.
        subscribe_to_facts: list = [
                                       "cognito-ehelply",

                                       "db-rds-ehelply",

                                       "mq-logging",
                                       "mq-monitoring",
                                       "mq-usage",

                                       "ehelply-templates",
                                       "ehelply-sam"
                                   ] + self.get_fact_subscriptions()

        active_integrations: List[
            str] = self.get_service_required_integrations() + self.get_service_optional_integrations()

        if CONST_INTEGRATION_ACCESS in active_integrations:
            subscribe_to_facts.append("ehelply-access")

        if CONST_INTEGRATION_SECURITY in active_integrations:
            subscribe_to_facts.append("ehelply-security")

        if CONST_INTEGRATION_NOTES in active_integrations:
            subscribe_to_facts.append("ehelply-notes")

        if CONST_INTEGRATION_META in active_integrations:
            subscribe_to_facts.append("ehelply-meta")

        if CONST_INTEGRATION_USERS in active_integrations:
            subscribe_to_facts.append("ehelply-users")

        if CONST_INTEGRATION_PRODUCTS in active_integrations:
            subscribe_to_facts.append("ehelply-products")

        if CONST_INTEGRATION_NOTIFICATIONS in active_integrations:
            subscribe_to_facts.append("mq-notifications")
            subscribe_to_facts.append("ehelply-notifications")

        if CONST_INTEGRATION_PLACES in active_integrations:
            pass
            # TODO: Once added to facts: subscribe_to_facts.append("ehelply-places")

        if CONST_INTEGRATION_REVIEWS in active_integrations:
            pass
            # TODO: Once added to facts: subscribe_to_facts.append("ehelply-reviews")

        if CONST_INTEGRATION_APPOINTMENTS in active_integrations:
            pass
            # TODO: Once added to facts: subscribe_to_facts.append("ehelply-appointments")

        # Register the fact service integration
        self.register_integration(Fact(State.logger, subscribe_to_facts))

        # Register the M2M service integration
        self.register_integration(M2M(self.service_gatekeeper_key))

        # Register the SDK service integration
        self.register_integration(SDK(self.service_gatekeeper_key))

        # Register the core service integration
        self.register_integration(Log(self.service_meta.key))
        self.register_integration(Monitor())

        # Register the service integrations
        if CONST_INTEGRATION_NOTIFICATIONS in active_integrations:
            self.register_integration(Notifications(self.get_notification_types()))

    def get_mysql_credentials(self) -> MySQLCredentials:
        """
        Returns MySQL credentials

        Returns:
            MySQLCredentials: SQL Credentials

        """
        # rds_fact: dict = get_fact_stage("db-rds-ehelply")
        db_secret = State.secrets.get(category=secret_name_database())[0]
        return MySQLCredentials(
            host=db_secret.host,
            database=db_secret.dbname,
            port=db_secret.port,
            username=db_secret.username,
            password=db_secret.password,
        )

    def facts(self):
        """
        Push all the fact providers to the fact service

        Returns:
            None

        """
        if len(self.get_fact_providers()) > 0:
            State.logger.info("Pushing facts...")
        for fact in self.get_fact_providers():
            fact_integration: Fact = State.integrations.get("fact")
            fact_integration.push_fact(fact)

    def migrate(self):
        """
        Preforms MySQL Migrations

        Returns:
            None

        """
        if LOADABLE_MYSQL in self.service_loadables and State.config.sql.migrate:
            State.logger.info("Running Migrations..")
            import alembic.config
            alembicArgs = [
                '--raiseerr',
                'upgrade', 'head',
            ]
            alembic.config.main(argv=alembicArgs)

    async def monitor(self):
        """
        Creates link to a monitor service via a new thread

        Returns:
            None

        """
        monitor_integration: Monitor = State.integrations.get("monitor")
        # monitor_integration.register_service()

        State.logger.info("Booting async monitor... Expect to see out of order Batcher info in the logs.")
        from ehelply_microservice_library.routers.router_service_monitor import MonitorHeartbeatThread
        monitor_thread = MonitorHeartbeatThread(
            logger=State.logger,
            heartbeat_interval_minutes=get_fact_stage("mq-monitoring")["interval"],
            queue_url=get_fact_endpoint("mq-monitoring"),
            environment=Environment.stage(),
            service_meta=self.service_meta,
            service_status=self.service_status,
        )
        monitor_thread.start()

    def check_for_updates(self):
        """
        Checks for service template updates

        Returns:
            None

        """
        access_key: str = State.secrets.get(category=secret_name_updater_vault())[0].access
        secret_key: str = State.secrets.get(category=secret_name_updater_vault())[0].secret

        service_updater: ServiceUpdater = ServiceUpdater(access_key=access_key, secret_key=secret_key,
                                                         root_path=self.get_service_package_path())
        # update_available: bool = service_updater.is_update_available()

        try:
            update_available: bool = service_updater.is_update_available()
            if update_available:
                State.logger.info("New updates are available for the service template")
                State.logger.info("Service template update preview: " + str(service_updater.preview().dict()))
            else:
                State.logger.info("No service template updates are available")
        except:
            State.logger.warning("Unable to check for service template updates")

    async def fetch_keys(self):
        """
        Fetches keys from the security microservice

        Returns:
            None

        """
        from ehelply_microservice_library.routers.router_service_security import refresh_encryption_keys
        await refresh_encryption_keys(category=self.service_meta.key)

        State.logger.info("Booting security key fetcher... Expect to see out of order Batcher/Timer info in the logs.")
        from ehelply_microservice_library.routers.router_service_security import SecurityKeysThread
        security_keys_thread = SecurityKeysThread(
            logger=State.logger,
            delay=20 * 60,  # 20 min
            category=self.service_meta.key
        )
        security_keys_thread.start()

    async def fetch_kpis(self):
        """
        Fetches KPIs from this microservice. Look into bootstrapper.get_kpi()

        Returns:
            None

        """
        State.logger.info("Booting service KPI fetcher... Expect to see out of order Batcher/Timer info in the logs.")
        from ehelply_microservice_library.routers.router_service_monitor import KPIFetcherThread
        kpi_fetcher_thread = KPIFetcherThread(
            logger=State.logger,
            queue_url=get_fact_endpoint("mq-monitoring"),
            service_meta=self.service_meta
        )
        kpi_fetcher_thread.start()

    def configure_security(self):
        """
        Hashing cost factor is a function of 2^COST_FACTOR

        Returns:
            None

        """
        Hashing.COST_FACTOR = 12

    def before_boot(self):
        """
        EVENT HOOK - Runs before microservice begins booting

        Returns:
            None

        """
        State.logger.info("Setting up security")

        self.configure_security()

    async def before_integrations(self):
        """
        EVENT HOOK - Runs before microservice loads integrations

        Returns:
            None

        """
        State.logger.info("Fetching secrets...")

        self.fetch_secrets()

    async def before_drivers(self):
        """
        EVENT HOOK - Runs before microservice loads drivers

        Returns:
            None

        """
        State.logger.info("Pushing service specific facts...")

        self.facts()

        State.logger.info("Fetching encryption keys...")

        await self.fetch_keys()

    async def post_load(self):
        """
        Entry point of the application that is executed after the bootstrapping process completes

        If the microservice operates completely based on FastAPI endpoints, you can leave this method as is.
        If you require any custom logic or additional threads to be spun up, do it in this method.

        Returns:
            None

        """
        State.logger.info("Setting up Dynamo auth")

        AuthDynamos.make()

        State.logger.info("Integrating into cluster...")

        await self.monitor()

        State.logger.info("Setting up KPI fetching...")

        await self.fetch_kpis()

        State.logger.info("Checking for service template updates...")

        self.check_for_updates()

        State.logger.info("Running application code...")
